@@ -14,7 +14,7 @@
  * limitations under the License.
  */
 
-/* @version $Id: jsvc-unix.c 1023277 2010-10-16 13:08:31Z mturk $ */
+/* @version $Id: jsvc-unix.c 1130635 2011-06-02 16:29:44Z mturk $ */
 #include "jsvc.h"
 
 #include <signal.h>
@@ -57,9 +57,11 @@ pid_t logger_pid = 0;           /* the logger process pid */
 static bool stopping = false;
 static bool doreload = false;
 static bool doreopen = false;
+static bool dosignal = false;
 typedef void (*sighandler_t)(int);
 static sighandler_t handler_int  = NULL;
 static sighandler_t handler_usr1 = NULL;
+static sighandler_t handler_usr2 = NULL;
 static sighandler_t handler_hup  = NULL;
 static sighandler_t handler_trm  = NULL;
 
@@ -129,6 +131,10 @@ static void handler(int sig)
         case SIGUSR1:
              log_debug("Caught SIGUSR1: Reopening logs");
              doreopen = true;
+	break;
+        case SIGUSR2:
+	     log_debug("Caught SIGUSR2: Scheduling a custom signal");
+             dosignal = true;
         break;
         default:
             log_debug("Caught unknown signal %d", sig);
@@ -374,6 +380,7 @@ static void controller(int sig)
         case SIGINT:
         case SIGHUP:
         case SIGUSR1:
+        case SIGUSR2:
             log_debug("Forwarding signal %d to process %d", sig, controlled);
             kill(controlled, sig);
             signal(sig, controller);
@@ -658,7 +665,7 @@ static int child(arg_data *args, home_data *data, uid_t uid, gid_t gid)
     /* Check wether we need to dump the VM version */
     if (args->vers == true) {
         log_error("jsvc (Apache Commons Daemon) " JSVC_VERSION_STRING);
-        log_error("Copyright (c) 1999-2010 Apache Software Foundation.");
+        log_error("Copyright (c) 1999-2011 Apache Software Foundation.");
         if (java_version() != true) {
             return -1;
         }
@@ -711,6 +718,7 @@ static int child(arg_data *args, home_data *data, uid_t uid, gid_t gid)
     /* Install signal handlers */
     handler_hup = signal_set(SIGHUP, handler);
     handler_usr1 = signal_set(SIGUSR1, handler);
+    handler_usr2 = signal_set(SIGUSR2, handler);
     handler_trm = signal_set(SIGTERM, handler);
     handler_int = signal_set(SIGINT, handler);
     controlled = getpid();
@@ -729,6 +737,10 @@ static int child(arg_data *args, home_data *data, uid_t uid, gid_t gid)
             doreopen = false;
             set_output(args->outfile, args->errfile, args->redirectstdin, args->procname);
         }
+        if(dosignal) {
+            dosignal = false;
+            java_signal();
+        }
     }
     remove_tmp_file(args);
     log_debug("Shutdown or reload requested: exiting");
@@ -777,47 +789,77 @@ static int logger_child(int out_fd, int err_fd, char *procname)
 {
     fd_set rfds;
     struct timeval tv;
-    int retval, n;
+    int retval, nfd = -1, rc = 0;
+    ssize_t n;
     char buf[LOGBUF_SIZE];
 
-    if (out_fd > err_fd) {
-        n = out_fd + 1;
-    } else {
-        n = err_fd + 1;
-    }
+    if (out_fd == -1 && err_fd == -1)
+        return EINVAL;
+    if (out_fd == -1)
+        nfd = err_fd;
+    else if (err_fd == -1)
+        nfd = out_fd;
+    else
+        nfd = out_fd > err_fd ? out_fd : err_fd;
+    ++nfd;
 
     openlog(procname, LOG_PID, LOG_DAEMON);
 
-    while (1) {
+    while (out_fd != -1 || err_fd != -1) {
         FD_ZERO(&rfds);
-        FD_SET(out_fd, &rfds);
-        FD_SET(err_fd, &rfds);
-        tv.tv_sec = 60;
+        if (out_fd != -1) {
+            FD_SET(out_fd, &rfds);
+        }
+        if (err_fd != -1) {
+            FD_SET(err_fd, &rfds);
+        }
+        tv.tv_sec  = 60;
         tv.tv_usec = 0;
-        retval = select(n, &rfds, NULL, NULL, &tv);
-        if (retval == -1)
+        retval = select(nfd, &rfds, NULL, NULL, &tv);
+        if (retval == -1) {
+            rc = errno;
             syslog(LOG_ERR, "select: %s", strerror(errno));
+            /* If select failed no point to continue */
+            break;
+        }
         else if (retval) {
-            if (FD_ISSET(out_fd, &rfds)) {
-                ssize_t n = read(out_fd, buf, LOGBUF_SIZE-1);
-                if (n < 0) {
+            if (out_fd != -1 && FD_ISSET(out_fd, &rfds)) {
+                do {
+                    n = read(out_fd, buf, LOGBUF_SIZE-1);
+                } while (n == -1 && errno == EINTR);
+                if (n == -1) {
                     syslog(LOG_ERR, "read: %s", strerror(errno));
-                } else if (n > 0 && buf[0] != '\n') {
+                    close(out_fd);
+                    if (err_fd == -1)
+                        break;
+                    nfd = err_fd + 1;
+                    out_fd = -1;
+                }
+                else if (n > 0 && buf[0] != '\n') {
                     buf[n] = 0;
                     syslog(LOG_INFO, "%s", buf);
                 }
             }
-            if (FD_ISSET(err_fd, &rfds)) {
-                ssize_t n = read(err_fd, buf, LOGBUF_SIZE-1);
-                if (n < 0) {
+            if (err_fd != -1 && FD_ISSET(err_fd, &rfds)) {
+                do {
+                    n = read(err_fd, buf, LOGBUF_SIZE-1);
+                } while (n == -1 && errno == EINTR);
+                if (n == -1) {
                     syslog(LOG_ERR, "read: %s", strerror(errno));
-                } else if (n > 0 && buf[0] != '\n') {
+                    close(err_fd);
+                    if (out_fd == -1)
+                        break;
+                    nfd = out_fd + 1;
+                    err_fd = -1;
+                }
+                else if (n > 0 && buf[0] != '\n') {
                     buf[n] = 0;
                     syslog(LOG_ERR, "%s", buf);
                 }
             }
         }
     }
+    return rc;
 }
 
 /**
@@ -825,7 +867,9 @@ static int logger_child(int out_fd, int err_fd, char *procname)
  */
 static void set_output(char *outfile, char *errfile, bool redirectstdin, char *procname)
 {
-    int out_pipe[2] = {0, 0}, err_pipe[2] = {0, 0}, fork_needed = 0;
+    int out_pipe[2] = {-1, -1};
+    int err_pipe[2] = {-1, -1};
+    int fork_needed = 0;
 
     if (redirectstdin == true) {
         freopen("/dev/null", "r", stdin);
@@ -896,24 +940,29 @@ static void set_output(char *outfile, char *errfile, bool redirectstdin, char *p
         pid_t pid = fork();
         if (pid == -1) {
             log_error("cannot create logger process: %s", strerror(errno));
-        } else {
-            if (pid) {
+        }
+        else {
+            if (pid != 0) {
+                /* Parent process.
+                 * Close child pipe endpoints.
+                 */
                 logger_pid = pid;
-                if (out_pipe[0] != 0) {
+                if (out_pipe[0] != -1) {
                     close(out_pipe[0]);
                     if (dup2(out_pipe[1], 1) == -1) {
                         log_error("cannot redirect stdout to pipe for syslog: %s",
                                   strerror(errno));
                     }
                 }
-                if (err_pipe[0] != 0) {
+                if (err_pipe[0] != -1) {
                     close(err_pipe[0]);
                     if (dup2(err_pipe[1], 2) == -1) {
                         log_error("cannot redirect stderr to pipe for syslog: %s",
                                   strerror(errno));
                     }
                 }
-            } else {
+            }
+            else {
                 exit(logger_child(out_pipe[0], err_pipe[0], procname));
             }
         }
@@ -1073,6 +1122,7 @@ static int run_controller(arg_data *args, home_data *data, uid_t uid,
 #endif
         signal(SIGHUP, controller);
         signal(SIGUSR1, controller);
+        signal(SIGUSR2, controller);
         signal(SIGTERM, controller);
         signal(SIGINT, controller);
 
